# -*- coding: utf-8 -*-

# Standard Library
import datetime
from dateutil.relativedelta import relativedelta

# Vendor Library
from django.db.models import Q
from m3.actions import ApplicationLogicException

# Project Library
from kinder.core.dict.models import UnitKind
from kinder.core.unit.models import Unit
from gisdo.constants import (
    AGE_CATEGORIES_FULL,
    AGE_CATEGORIES_EIGHT,
    AGE_CATEGORIES_CUT, ALL)
from gisdo import settings as gs_settings


class ViewList(object):

    def __init__(self):
        self._data = []
        self._index = 0

    def append(self, tpl):

        description, value = tpl

        data_value = (self._index, description, value)

        self._data.append(data_value)
        self._index += 1

    def get_data(self):

        return self._data


class DataDictionary(object):
    """
    Словарь, который содержит в себе индекс
    """

    def __init__(self):

        self._data_dictionary = {}
        self.index = 0

    def __setitem__(self, key, value):

        # Пока simple_report не видит параметр,
        # который внутри себя содержит символы "-", "."
        key = key.replace('-', '_')
        key = key.replace('.', '_')
        self._data_dictionary[key] = value
        self.index += 1

    def get_data(self):

        return self._data_dictionary


class DateMixin(object):
    """
    Класс-примесь добавляет к классу ряд
    методов для вычисления различных дат
    """

    LAST_ACADEMIC_DAY = 1
    LAST_ACADEMIC_MONTH = 9

    @classmethod
    def get_desired_date(cls):
        """
        Последний день текущего учебного
        года LAST_ACADEMIC_DAY.LAST_ACADEMIC_MONTH.год
        """

        return cls._get_current_calendar_year()

    @classmethod
    def _get_next_calendar_year(cls):
        """
        Возвращает текущий календарный год
        """

        today = datetime.date.today()
        today_month = today.month

        if today_month < 9:
            last_day = datetime.date(
                today.year, cls.LAST_ACADEMIC_MONTH,
                cls.LAST_ACADEMIC_DAY)
        else:
            last_day = datetime.date(
                today.year + 1, cls.LAST_ACADEMIC_MONTH,
                cls.LAST_ACADEMIC_DAY)

        return last_day

    @classmethod
    def _get_current_calendar_year(cls):
        """
        Возвращает текущий календарный год
        """

        today = datetime.date.today()
        today_month = today.month

        if today_month < 9:
            last_day = datetime.date(
                today.year - 1, cls.LAST_ACADEMIC_MONTH,
                cls.LAST_ACADEMIC_DAY)
        else:
            last_day = datetime.date(
                today.year, cls.LAST_ACADEMIC_MONTH,
                cls.LAST_ACADEMIC_DAY)

        return last_day

    @staticmethod
    def get_report_period():

        today = datetime.date.today()

        beg = datetime.date(today.year, 1, 1)
        end = datetime.date(today.year, today.month, today.day)

        return beg, end

    @classmethod
    def get_desired_period(cls):
        pass


class AgeDeltas(object):

    ACADEMIC_YEAR = 1
    CALENDAR_YEAR = 0

    _map = {
        # код: (количество лет, месяцев)
        '2_months': (0, 0),
        '6_months': (0, 6),
        '1_year': (1, 0),
        '1.5_year': (1, 6),
        '2_year': (2, 0),
        '2.5_year': (2, 6),
        '3_year': (3, 0),
        '3.5_year': (3, 6),
        '4_year': (4, 0),
        '4.5_year': (4, 6),
        '5_year': (5, 0),
        '5.5_year': (5, 6),
        '6_year': (6, 0),
        '6.5_year': (6, 6),
        '7_year': (7, 0),
        '7.5_year': (7, 6)
    }

    _plane_category_map = {
        ALL: ('7.5_year', '2_months'),
        AGE_CATEGORIES_FULL['2-6-MONTHS']: ('6_months', '2_months'),
        AGE_CATEGORIES_FULL['0.5-1-YEARS']: ('1_year', '6_months'),
        AGE_CATEGORIES_FULL['1-1.5-YEARS']: ('1.5_year', '1_year'),
        AGE_CATEGORIES_FULL['1.5-2-YEARS']: ('2_year', '1.5_year'),
        AGE_CATEGORIES_FULL['2-2.5-YEARS']: ('2.5_year', '2_year'),
        AGE_CATEGORIES_FULL['2.5-3-YEARS']: ('3_year', '2.5_year'),
        AGE_CATEGORIES_FULL['3-3.5-YEARS']: ('3.5_year', '3_year'),
        AGE_CATEGORIES_FULL['3.5-4-YEARS']: ('4_year', '3.5_year'),
        AGE_CATEGORIES_FULL['4-4.5-YEARS']: ('4.5_year', '4_year'),
        AGE_CATEGORIES_FULL['4.5-5-YEARS']: ('5_year', '4.5_year'),
        AGE_CATEGORIES_FULL['5-5.5-YEARS']: ('5.5_year', '5_year'),
        AGE_CATEGORIES_FULL['5.5-6-YEARS']: ('6_year', '5.5_year'),
        AGE_CATEGORIES_FULL['6-6.5-YEARS']: ('6.5_year', '6_year'),
        AGE_CATEGORIES_FULL['6.5-7-YEARS']: ('7_year', '6.5_year'),
        AGE_CATEGORIES_FULL['7-7.5-YEARS']: ('7.5_year', '7_year'),
    }

    _plane_eight_category_map = {
        ALL: ('7.5_year', '2_months'),
        AGE_CATEGORIES_EIGHT['2-1-YEARS']: ('1_year', '2_months'),
        AGE_CATEGORIES_EIGHT['1-2-YEARS']: ('2_year', '1_year'),
        AGE_CATEGORIES_EIGHT['2-3-YEARS']: ('3_year', '2_year'),
        AGE_CATEGORIES_EIGHT['3-4-YEARS']: ('4_year', '3_year'),
        AGE_CATEGORIES_EIGHT['4-5-YEARS']: ('5_year', '4_year'),
        AGE_CATEGORIES_EIGHT['5-6-YEARS']: ('6_year', '5_year'),
        AGE_CATEGORIES_EIGHT['6-7-YEARS']: ('7_year', '6_year'),
        AGE_CATEGORIES_EIGHT['7-7.5-YEARS']: ('7.5_year', '7_year')
    }

    _plane_cut_category_map = {
        AGE_CATEGORIES_CUT['2M-3YEARS']: ('3_year', '2_months'),
        AGE_CATEGORIES_CUT['3-5-YEARS']: ('5_year', '3_year'),
        AGE_CATEGORIES_CUT['5-7.5-YEARS']: ('7.5_year', '5_year'),
    }

    @classmethod
    def _get_age_deltas(cls, date, age_code):

        if callable(date):
            d = date()
        else:
            d = date

        years, months = cls._map[age_code]
        return d - relativedelta(years=years, months=months)

    @staticmethod
    def calculate_date():

        today = datetime.date.today()

        def calculate_academic_date():
            """
            Способ расчета возраста на текущий учебный год
            """
            return (datetime.date(today.year, 9, 1)
                    if today >= datetime.date(today.year, 9, 1)
                    else datetime.date(today.year - 1, 9, 1))

        def calculate_calendar_date():
            """
            Способ расчета возраста на текущий календарный год
            """
            return datetime.date(today.year, 9, 1)

        if gs_settings.AGE_CALCULATION_DATE == AgeDeltas.CALENDAR_YEAR:
            date = calculate_calendar_date()
        elif gs_settings.AGE_CALCULATION_DATE == AgeDeltas.ACADEMIC_YEAR:
            date = calculate_academic_date()
        else:
            raise ApplicationLogicException(
                u'В конфигурациях неверно задан способ '
                u'расчета возраста ребенка')

        return date

    @classmethod
    def get_category_deltas(cls, category_code, date_func):

        if category_code in cls._plane_category_map:
            down_border, up_border = cls._plane_category_map[category_code]
        elif category_code in cls._plane_cut_category_map:
            down_border, up_border = cls._plane_cut_category_map[
                category_code]
        else:
            down_border, up_border = cls._plane_eight_category_map[
                category_code]

        return (
            cls._get_age_deltas(date_func, down_border),
            cls._get_age_deltas(date_func, up_border))


class UnitHelper(object):
    """
    Класс для работы с учреждениями
    """

    def __init__(self, report_main_unit):

        self._unit = report_main_unit
        self.is_region = (self._unit.kind.id == UnitKind.REGION)

        # Муниципальные учреждения по которым собирается отчет
        self._mo_iterable = self._get_high_level_units()

        self._dou_units = dict([
            (mo.id, [r for r in self._get_list_dou(mo)])
            for mo in self._mo_iterable])

        # Общее количество учреждений по которым собирается отчет
        self._unit_count = self._unit.get_descendants().filter(
            kind__id__in=
            [UnitKind.DOU,
             UnitKind.MO,
             UnitKind.REGION]).count()

    def _get_high_level_units(self):

        return Unit.objects.raw(
            'select * '
            'from unit '
            'join (select unit.id, count(t1.id) as count_1'
            '      from unit'
            '      join (select id, lft, rght, tree_id'
            '            from unit'
            '            where kind_id=%s) as t1'
            '      on t1.lft >= unit.lft and'
            '               t1.rght <= unit.rght and'
            '               t1.tree_id = unit.tree_id'
            '      where unit.kind_id = %s'
            '      group by unit.id) as t2 '
            'on t2.id = unit.id '
            'where t2.count_1 > 0 and unit.lft >= %s and'
            '                         unit.rght <= %s and'
            '                         unit.tree_id = %s' %
            (UnitKind.DOU, UnitKind.MO,
             self._unit.lft, self._unit.rght,
             self._unit.tree_id))

    @staticmethod
    def _get_list_dou(unit):

        if unit.kind.id == UnitKind.MO:

            not_ready_dou_filter = Q(
                gisdo__not_on_federal_report=False
            ) & Q(build_in_progress=False)
            result = (
                unit.get_descendants(
                    include_self=True).filter(
                        kind__id=UnitKind.DOU)
                .filter(not_ready_dou_filter).order_by('kind__id'))
        else:
            result = [unit]

        return result

    def get_main_unit(self):

        return self._unit

    def get_report_units(self):
        """
        Муниципальные учреждения по которым собирается ФО
        """

        return self._mo_iterable

    def get_mo_units(self, mo):
        """
        Все сады заданного МО
        """

        return self._dou_units[mo] if mo in self._dou_units else []
